﻿//*************************************************************************************************
//【サンプルプログラム】『Kinectのデプスイメージから呼吸をモニターする』
//【名称】KinectVitalMonitor　【Version】0.01
//【作成者】私設研究所Neo-Tech-Lab.com  上田智章
//【作成日】2012年10月15日　【コメント追記】2012年10月15日
//*************************************************************************************************

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;             //【Debug】を使うための定義
using System.Drawing;                 //【Bitmap】【Color】を使うための定義
using System.Drawing.Imaging;         //【BitmapData】を使うための定義
using System.Runtime.InteropServices; //【Marshal.Copy】を使うための定義
using Microsoft.Kinect;               //【参照設定】プロジェクト(P)⇒参照の追加(F)⇒.NETタブ⇒Microsoft.Kinectを選択
using NTL.KinectLib;                  //【NTL_KinectLib】ColorBar

namespace KinectVitalMonitor
{
    public partial class Form1 : Form
    {
        //【KINECT関連】
        //　関数内でダイナミックに変数を確保すると動作速度が遅くなるので、ここにまとめて変数を確保しています。
        private KinectSensor KS_KinectSensor;   //【KINECT】
        private bool KS_flagFlip = true;        //【左右反転フラグ】
        private const int KS_FrameWidth = 640;  //【KINECTのFrameの横方向画素数】
        private const int KS_FrameHeight = 480; //【KINECTのFrameの縦方向画素数】
        private Rectangle KS_rect;              //【Image表示枠】

        private byte[] KS_IrCamData;            //【KINECTのIrCamImageData】16bit/画素
        private byte[] KS_ColorImageDataXRGB;   //【KINECTのColorImageData】32bit/画素,X8R8G8B8表示形式
        private Bitmap KS_bmp1;                 //【ColorImage表示用Bitmapオブジェクト】
        private BitmapData KS_data1;            //【ColorImage用BitmapData】
        private int IrMin = 3000;               //【赤外線強度最小値】
        private int IrMax = 65000;              //【赤外線強度最大値】
        private int capIrTime;                  //【心拍数演算用時刻】

        private DepthImagePixel[] KS_DepthData; //【KINECTのDepthImageData】Depth, PlayerIndex 
        private int KS_Lmin;                    //【深度測定可能最近距離】default:  800mm, near: 400mm 
        private int KS_Lmax;                    //【深度測定可能最遠距離】default: 4000mm, near:3000mm
        private ColorImagePoint[] KS_ScreenPos; //【KINECTのDepthImageData】16bit/画素(MSB13bit:Depth, LSB3bit:PlayerID) 
        private byte[] KS_DepthImageDataXRGB;   //【KINECTのDepthImageData】32bit/画素,X8R8G8B8表示形式
        private Bitmap KS_bmp2;                 //【DepthImage表示用Bitmapオブジェクト】
        private BitmapData KS_data2;            //【DepthImage用BitmapData】

        private Skeleton[] KS_SkeletonData;     //【骨格情報】
        private SkeletonPoint[] KS_SkeletonPoint = new SkeletonPoint[6];
        private ColorImagePoint[] KS_ColorPoint = new ColorImagePoint[6];

        private Rectangle RectFace;   //【顔】
        private int IrCount = 0;         //【計測回数】
     // private RectangleWaveFilter IrFilter = new RectangleWaveFilter();
        private HeartRateFilter HeartFilter = new HeartRateFilter();
        private LowPassFilter IrLPF = new LowPassFilter();
        private FIFO IrFIFO = new FIFO();
        private NoiseReduction IrNRFilter = new NoiseReduction();

        private Rectangle RectBreast; //【胸部】
        private int BrCount = 0;         //【デプス計測回数】
        private FIFO BrFIFO = new FIFO();
        private NoiseReduction BrNRFilter = new NoiseReduction();

        private const int nFIFO = 601;   //【波形表示の為のデータ格納配列サイズ】
        private const int nOffset = 150; //【オフセット除去処理サンプル数】
        private const int nWave = 24;    //【矩形波相関フィルタ処理サンプル数】★必ず4の倍数
        private const int nLPF = 12;     //【赤外線脈波のLPF処理サンプル数】
        private const int nBrLPF = 30;   //【呼吸ノイズ除去処理サンプル数】

        private double BrValue = 0.0;  //【単位mm】【呼吸値】
        private double BrOffset = 0.0; //【単位mm】【呼吸オフセット】
        private double IrValue = 0.0;  //【単位digit】【脈波値】
        private double IrOffset = 0.0; //【単位digit】【脈波オフセット】

        private Boolean fBreathing = true;
        private Boolean fHeartRate = true;


        //【NTL拡張現実センサ関連】---------------------------------------------------------
        private ColorBar ColorBar1 = new ColorBar(); //【赤外線強度をカラー値に変換する部品】
        private ColorBar ColorBar2 = new ColorBar(); //【デプス値をカラー値に変換する部品】

        public Form1()
        {
            InitializeComponent();
            //***********************************************************
            //【Kinect検出】【共通】
            //***********************************************************
            //【利用可能なKINECTセンサを取得する】
            KS_KinectSensor = (from sensorToCheck
                               in KinectSensor.KinectSensors
                               where sensorToCheck.Status == KinectStatus.Connected
                               select sensorToCheck).FirstOrDefault();

            //【検出結果の表示】
            pictureBox1.SizeMode = PictureBoxSizeMode.StretchImage; //【デプス画像をpictureBox1全体に引き伸ばして表示する】
            pictureBox2.SizeMode = PictureBoxSizeMode.StretchImage;
            pictureBox3.SizeMode = PictureBoxSizeMode.StretchImage;
            pictureBox4.SizeMode = PictureBoxSizeMode.StretchImage;
            pictureBox2.Image = Bitmap.FromFile("../../image/KinectOff.png"); //【Kinect未検出の画像を表示】
            if (KS_KinectSensor == null)
            { //【Kinect接続が確認できなかった場合】
                MessageBox.Show(
                   "接続可能なKINECTセンサが見当たりません。",
                   "KINECT for Windows Error Message",
                   MessageBoxButtons.OK
                );
                return;
            }
            //【Kinect検出に成功】
            pictureBox2.Image = Bitmap.FromFile("../../image/KinectOn.png"); //【Kinectセンサ画像を表示】
            pictureBox3.Image = Bitmap.FromFile("../../image/ColorBar.jpg"); //【Kinectセンサ画像を表示】
            pictureBox4.Image = Bitmap.FromFile("../../image/ColorBar.jpg"); //【Kinectセンサ画像を表示】

            //***********************************************************
            //【Kinectの初期化処理】
            //***********************************************************
            //【KINECTの設定】DepthStreamを許可
            KS_KinectSensor.ColorStream.Enable(ColorImageFormat.InfraredResolution640x480Fps30); //【ColorStream許可】
            KS_KinectSensor.DepthStream.Enable(DepthImageFormat.Resolution640x480Fps30);         //【DepthStream許可】
            KS_KinectSensor.SkeletonStream.Enable();                                             //【SkeletonStream許可】
            //【KINECT動作開始】
            KS_KinectSensor.Start(); //【KINECTスタート】

            //【KINECTデータ取得イベントの設定とデータバッファの確保】
            KS_KinectSensor.ColorFrameReady += new EventHandler<ColorImageFrameReadyEventArgs>(KS_ColorFrameReady);
            KS_KinectSensor.DepthFrameReady += new EventHandler<DepthImageFrameReadyEventArgs>(KS_DepthFrameReady);
            KS_KinectSensor.SkeletonFrameReady += new EventHandler<SkeletonFrameReadyEventArgs>(KS_SkeletonFrameReady);
                //.AllFramesReady += new EventHandler<AllFramesReadyEventArgs>(KS_KinectSensor_AllFramesReady);

            //【Form終了イベント】
            this.FormClosing += new FormClosingEventHandler(Form1_FormClosing);
            radioButton1.CheckedChanged += new EventHandler(DepthRangeChange);
            radioButton2.CheckedChanged += new EventHandler(DepthRangeChange);
            checkBox1.CheckedChanged += new EventHandler(checkBox1_CheckedChanged);

            KS_IrCamData = new byte[2 * KS_FrameWidth * KS_FrameHeight];          //【赤外線強度データ１フレーム分を保存するメモリ】
            KS_ColorImageDataXRGB = new byte[4 * KS_FrameWidth * KS_FrameHeight]; //【赤外線画像表示用】
            KS_DepthData = new DepthImagePixel[KS_FrameWidth * KS_FrameHeight];   //【深度情報データ１フレーム分を保存するメモリを確保】
            KS_DepthImageDataXRGB = new byte[4 * KS_FrameWidth * KS_FrameHeight]; //【デプス画像表示用】
            KS_ScreenPos = new ColorImagePoint[KS_FrameWidth * KS_FrameHeight];   //【座標情報】
            KS_SkeletonData = new Skeleton[KS_KinectSensor.SkeletonStream.FrameSkeletonArrayLength]; //【骨格情報用】


            KS_rect = new Rectangle(0, 0, KS_FrameWidth, KS_FrameHeight); //【DepthImage用枠】
            KS_bmp1 = new Bitmap(KS_FrameWidth, KS_FrameHeight, System.Drawing.Imaging.PixelFormat.Format32bppRgb);  //【ColorImage表示用】
            KS_bmp2 = new Bitmap(KS_FrameWidth, KS_FrameHeight, System.Drawing.Imaging.PixelFormat.Format32bppArgb); //【DepthImage表示用】

            //
            if (checkBox1.Checked) { KS_flagFlip = true; } else { KS_flagFlip = false; }

            KS_KinectSensor.DepthStream.Range = DepthRange.Default;    //【Kinect for Windows】【Depth測定モードをDefaultに設定】
            label8.Text = "Default";
            KS_CheckDistance();                                     //【設定結果を表示】Lmax, Lminを取得

            KS_KinectSensor.CoordinateMapper.MapDepthFrameToColorFrame(
                DepthImageFormat.Resolution640x480Fps30,
                KS_DepthData,
                ColorImageFormat.InfraredResolution640x480Fps30,
                KS_ScreenPos);

            //【Kinectヘッドの仰角を設定】設定範囲は±27°
            int HeadAngle = -10;
            KS_KinectSensor.ElevationAngle = HeadAngle; //【ベッドより上方から撮影するのでヘッドを下に向ける】★要調整
            textBox1.Text = HeadAngle.ToString();

            //【深度情報をカラーに変換する為の準備処理】
            ColorBar1.CreateColorPalletForInfrared();                      //【赤外線強度表示用カラーパレットを作成する】
            ColorBar2.CreateColorPalletForDistance(KS_Lmax + 1);           //【距離表示用カラーパレットを作成する】
            pictureBox3.Image = ColorBar1.CreateColorScale(256, 10, true); //【256×10画素サイズのカラーバーを作成して表示する】
            pictureBox4.Image = ColorBar2.CreateColorScale(256, 10, true); //【256×10画素サイズのカラーバーを作成して表示する】

        }


        void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            if (KS_KinectSensor != null)
            {
                KS_KinectSensor.Stop();    //【KINECTセンサを終了】★これを実行しないとレーザーが消えない
                KS_KinectSensor.Dispose(); //【KINECTセンサのオブジェクトを解放する】
            }
        }

        //***************************************************************
        //【距離モード変更】Defaultモード(800mm～4000mm) Nearモード(400m～3000mm)
        //***************************************************************
        void DepthRangeChange(object sender, EventArgs e)
        {
            if (radioButton1.Checked) { KS_KinectSensor.DepthStream.Range = DepthRange.Default; label8.Text = "Default"; }
            else { KS_KinectSensor.DepthStream.Range = DepthRange.Near; label8.Text = "Near"; }
            KS_CheckDistance();
        }

        //***************************************************************
        //【CheckDistance】KINECTセンサの検出可能距離現在値を調べる
        //***************************************************************
        private void KS_CheckDistance()
        {
            KS_Lmax = KS_KinectSensor.DepthStream.MaxDepth;
            KS_Lmin = KS_KinectSensor.DepthStream.MinDepth;
            label1.Text = "Min=" + KS_Lmin.ToString() + "mm";
            label2.Text = "Max=" + KS_Lmax.ToString() + "mm";
        }

        //***************************************************************
        //【上下角変更ボタン】KINECTの上下角を変更。±27°の範囲内。時間を置かないとエラーが発生する
        //***************************************************************
        private void button2_Click(object sender, EventArgs e)
        {
            Int32 HeadAngle = int.Parse(textBox1.Text); //【設定角度を読み込む】±27°
            if ((HeadAngle <= KS_KinectSensor.MaxElevationAngle) && (HeadAngle >= KS_KinectSensor.MinElevationAngle))
            {
                KS_KinectSensor.ElevationAngle = HeadAngle; //【上下角の変更】
            }
        }

        //***************************************************************
        //【カラー画像の左右反転指定フラグ】
        //***************************************************************
        private void checkBox1_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox1.Checked) { KS_flagFlip = true; } else { KS_flagFlip = false; }
        }

        //***************************************************************
        //【赤外線脈波表示フラグ】
        //***************************************************************
        private void checkBox2_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox2.Checked) { fHeartRate = true; } else { fHeartRate = false; }
        }

        //***************************************************************
        //【呼吸曲線表示フラグ】
        //***************************************************************
        private void checkBox3_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBox3.Checked) { fBreathing = true; } else { fBreathing = false; }
        }




        //***************************************************************
        //【IrCamImageの取得イベント】
        //***************************************************************
        private void KS_ColorFrameReady(object sender, ColorImageFrameReadyEventArgs e)
        {
            using (ColorImageFrame f = e.OpenColorImageFrame())
            {
                if (f != null)
                {
                    f.CopyPixelDataTo(KS_IrCamData);
                    capIrTime = System.Environment.TickCount; //【キャプチャー時刻を取得】
                    int i, j, k;
                    j = 0;
                    for (i = 0; i < KS_IrCamData.Length; i += 2)
                    {
                        k = KS_IrCamData[i + 1] * 256 + KS_IrCamData[i];
                        KS_ColorImageDataXRGB[j] = ColorBar1.B[k];
                        KS_ColorImageDataXRGB[j + 1] = ColorBar1.G[k];
                        KS_ColorImageDataXRGB[j + 2] = ColorBar1.R[k];
                        KS_ColorImageDataXRGB[j + 3] = ColorBar1.A[k];
                        j += 4;
                    }

                    KS_data1 = KS_bmp1.LockBits(KS_rect, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppRgb); //【Lock】
                    Marshal.Copy(KS_ColorImageDataXRGB, 0, KS_data1.Scan0, KS_ColorImageDataXRGB.Length); //【UnmanagedMemoryのBitmapDataに配列PixelDataをコピー】
                    KS_bmp1.UnlockBits(KS_data1);                                           //【Lock解除】
                    if (KS_flagFlip) { KS_bmp1.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
                    pictureBox1.BackgroundImage = KS_bmp1;                                  //【pictureBox1に画像表示】
                }
            }
        }

        //***************************************************************
        //【DepthImageの取得イベント】
        //***************************************************************
        private void KS_DepthFrameReady(object sender, DepthImageFrameReadyEventArgs e)
        {
            using (DepthImageFrame f = e.OpenDepthImageFrame())
            {
                if (f != null)
                {
                    f.CopyDepthImagePixelDataTo(KS_DepthData);
                    int i, j, k, p;
                    for (i = 0; i < KS_DepthData.Length; i++)
                    {
                        k = KS_DepthData[i].Depth;
                        if (k < 0) { k = 0; }
                        if (k >= KS_Lmax) { k = KS_Lmax; }
                        p = KS_DepthData[i].PlayerIndex;
                        j = 4 * (KS_ScreenPos[i].X + KS_FrameWidth * KS_ScreenPos[i].Y);
                        if (p > 0)
                        {
                            KS_DepthImageDataXRGB[j] = ColorBar2.B[k];
                            KS_DepthImageDataXRGB[j + 1] = ColorBar2.G[k];
                            KS_DepthImageDataXRGB[j + 2] = ColorBar2.R[k];
                            KS_DepthImageDataXRGB[j + 3] = 150; // ColorBar2.A[k];
                        }
                        else
                        {
                            KS_DepthImageDataXRGB[j] = 0;
                            KS_DepthImageDataXRGB[j + 1] = 0;
                            KS_DepthImageDataXRGB[j + 2] = 0;
                            KS_DepthImageDataXRGB[j + 3] = 0;
                        }
                        j += 4;
                    }

                    KS_data2 = KS_bmp2.LockBits(KS_rect, ImageLockMode.WriteOnly, System.Drawing.Imaging.PixelFormat.Format32bppArgb); //【Lock】
                    Marshal.Copy(KS_DepthImageDataXRGB, 0, KS_data2.Scan0, KS_ColorImageDataXRGB.Length); //【UnmanagedMemoryのBitmapDataに配列PixelDataをコピー】
                    KS_bmp2.UnlockBits(KS_data2);                                           //【Lock解除】
                    if (KS_flagFlip) { KS_bmp2.RotateFlip(RotateFlipType.Rotate180FlipY); } //【左右反転】
                    pictureBox1.Image = KS_bmp2;

                }
            }
        }

        //***************************************************************
        //【SkeletonFrameの取得イベント】SkeletonFrameは6名分。各々がJointsコレクションを持つ。
        //***************************************************************
        void KS_SkeletonFrameReady(object sender, SkeletonFrameReadyEventArgs e)
        {
            using (SkeletonFrame f = e.OpenSkeletonFrame())
            {
                if (f != null)
                {
                    f.CopySkeletonDataTo(KS_SkeletonData);
                    for (int i = 0; i < KS_SkeletonData.Length; i++)
                    {
                        if (KS_SkeletonData[i].TrackingState == SkeletonTrackingState.Tracked)
                        {
                            KS_SkeletonPoint[0] = KS_SkeletonData[i].Position;
                            KS_SkeletonPoint[1] = KS_SkeletonData[i].Joints[JointType.Head].Position;
                            KS_SkeletonPoint[2] = KS_SkeletonData[i].Joints[JointType.ShoulderCenter].Position;
                            KS_SkeletonPoint[3] = KS_SkeletonData[i].Joints[JointType.ShoulderLeft].Position;
                            KS_SkeletonPoint[4] = KS_SkeletonData[i].Joints[JointType.ShoulderRight].Position;
                            KS_SkeletonPoint[5] = KS_SkeletonData[i].Joints[JointType.HipCenter].Position;
                            if (KS_flagFlip)
                            {
                                KS_SkeletonPoint[0].X = -KS_SkeletonPoint[0].X;
                                KS_SkeletonPoint[1].X = -KS_SkeletonPoint[1].X;
                                KS_SkeletonPoint[2].X = -KS_SkeletonPoint[2].X;
                                KS_SkeletonPoint[3].X = -KS_SkeletonPoint[3].X;
                                KS_SkeletonPoint[4].X = -KS_SkeletonPoint[4].X;
                                KS_SkeletonPoint[5].X = -KS_SkeletonPoint[5].X;
                            }
                            KS_ColorPoint[0] = KS_KinectSensor.CoordinateMapper.MapSkeletonPointToColorPoint(KS_SkeletonPoint[0], ColorImageFormat.InfraredResolution640x480Fps30);
                            KS_ColorPoint[1] = KS_KinectSensor.CoordinateMapper.MapSkeletonPointToColorPoint(KS_SkeletonPoint[1], ColorImageFormat.InfraredResolution640x480Fps30);
                            KS_ColorPoint[2] = KS_KinectSensor.CoordinateMapper.MapSkeletonPointToColorPoint(KS_SkeletonPoint[2], ColorImageFormat.InfraredResolution640x480Fps30);
                            KS_ColorPoint[3] = KS_KinectSensor.CoordinateMapper.MapSkeletonPointToColorPoint(KS_SkeletonPoint[3], ColorImageFormat.InfraredResolution640x480Fps30);
                            KS_ColorPoint[4] = KS_KinectSensor.CoordinateMapper.MapSkeletonPointToColorPoint(KS_SkeletonPoint[4], ColorImageFormat.InfraredResolution640x480Fps30);
                            KS_ColorPoint[5] = KS_KinectSensor.CoordinateMapper.MapSkeletonPointToColorPoint(KS_SkeletonPoint[5], ColorImageFormat.InfraredResolution640x480Fps30);

                            int pX, pY, pW, pH; float Breathing, Heart;

                            //【顔皮膚検出領域】
                            pY = KS_ColorPoint[1].Y;
                            pH = KS_ColorPoint[2].Y - KS_ColorPoint[1].Y;
                            pW = pH;
                            pX = (int)((KS_ColorPoint[1].X + KS_ColorPoint[2].X - pW) * 0.5f);
                            if (pX < 0) { pX = 0; }
                            RectFace = new Rectangle(pX, pY, pW, pH);

                            if ((pW * pH < 30 * 30) || (pW < 0) || (pH < 0)) { return; }
                            if ((pX < 0) || (pY < 0)) { return; }

                            //【胸部検出領域】
                            pX = KS_ColorPoint[3].X;
                            pW = KS_ColorPoint[4].X - KS_ColorPoint[3].X;
                            if (KS_ColorPoint[3].Y < KS_ColorPoint[4].Y)
                            {
                                pY = KS_ColorPoint[3].Y;
                                pH = KS_ColorPoint[5].Y - KS_ColorPoint[3].Y;
                            } else {
                                pY = KS_ColorPoint[4].Y;
                                pH = KS_ColorPoint[5].Y - KS_ColorPoint[4].Y;
                            }
                            RectBreast = new Rectangle(pX, pY, pW, pH);

                            if ((pW * pH < 30 * 30) || (pW < 0) || (pH < 0)) { return; }
                            if ((pX < 0) || (pY < 0)) { return; }

                            //【呼吸関連処理】
                            Breathing = GetAverage(RectBreast, KS_Lmax, KS_Lmin);
                            if (BrCount == 0)
                            {
                                BrNRFilter.Initialize(nBrLPF);    //【最小二乗法でノイズ除去】【1秒分】
                                BrFIFO.Initialize(nFIFO, nOffset); //【波形表示20秒分,オフセット処理10秒分】
                            }
                            BrNRFilter.Estimation(Breathing); //【最小二乗法でノイズ除去】
                            BrValue = BrNRFilter.C;           //　：
                            BrFIFO.LoadDataToFIFO(BrValue);   //【波形表示用FIFOに格納】
                            BrOffset = BrFIFO.OffsetLevel;    //
                            BrCount++;

                            //【心拍関連処理】
                            if (IrCount == 0) //【初回はFIFOを初期化する】
                            {
                             //   IrFilter.Initialize(nWave);        //【矩形波相関フィルタの初期化】★4の倍数を指定すること
                                HeartFilter.Initialize();
                                IrNRFilter.Initialize(nWave);      //【最小２乗法でノイズ除去】
                            //  IrLPF.Initialize(nLPF);            //【LPFの初期化】
                                IrFIFO.Initialize(nFIFO, nOffset); //【波形表示20秒分,オフセット処理10秒分】
                            }
                            Heart = GetIrAverage(RectFace, IrMax, IrMin);        　//【距離測定】指定領域内の加算平均デプス値を求める
                          //  Heart = (float)IrFilter.CrossCorrelationFilter(Heart); //【矩形波相関フィルタ処理】
                            Heart = HeartFilter.GetFilter(Heart, capIrTime);                        
                            IrNRFilter.Estimation(Heart);                          //【高性能ノイズ除去フィルタ】
                            Heart = IrNRFilter.C;                                  //【高性能ノイズ除去フィルタ】
                            if (Heart < -8388608.0f) { return; } //*********************************************************
                            float HBR = HeartFilter.GetHeartBeatRate(Heart);
                        //  Heart = IrLPF.LPF(Heart);                              //【LPF】★上のフィルタで十分だったので未使用
                            IrFIFO.LoadDataToFIFO(Heart);                          //【波形表示用FIFOに格納】
                            IrValue = Heart;
                            IrOffset = BrFIFO.OffsetLevel;
                            IrCount++;

                            Graphics g = Graphics.FromImage(pictureBox1.Image);
                            g.DrawEllipse(new Pen(Color.SkyBlue, 3.0f), KS_ColorPoint[0].X - 5.0f, KS_ColorPoint[0].Y - 5.0f, 10.0f, 10.0f); //【Center】
                            g.DrawEllipse(new Pen(Color.Red, 3.0f), KS_ColorPoint[1].X - 5.0f, KS_ColorPoint[1].Y - 5.0f, 10.0f, 10.0f); //【Head】
                            g.DrawEllipse(new Pen(Color.Red, 3.0f), KS_ColorPoint[2].X - 5.0f, KS_ColorPoint[2].Y - 5.0f, 10.0f, 10.0f); //【ShoulderCenter】
                            g.DrawEllipse(new Pen(Color.Red, 3.0f), KS_ColorPoint[3].X - 5.0f, KS_ColorPoint[3].Y - 5.0f, 10.0f, 10.0f); //【ShoulderLeft】
                            g.DrawEllipse(new Pen(Color.Red, 3.0f), KS_ColorPoint[4].X - 5.0f, KS_ColorPoint[4].Y - 5.0f, 10.0f, 10.0f); //【ShoulderRight】
                            g.DrawEllipse(new Pen(Color.Red, 3.0f), KS_ColorPoint[5].X - 5.0f, KS_ColorPoint[5].Y - 5.0f, 10.0f, 10.0f); //【HipCenter】
                            g.DrawRectangle(new Pen(Color.Orange, 3.0f), RectFace);   //【測定枠】
                            g.DrawRectangle(new Pen(Color.Yellow, 3.0f), RectBreast); //【測定枠】

                            if (fHeartRate || fBreathing) { DisplayGrid(g, Color.LightPink); } //【グリッド表示】
                            if (fHeartRate) { DisplayHeartRateSignal(g, Color.HotPink); }      //【波形表示】赤外線反射強度⇒『心拍』
                            if (fBreathing) { DisplayBreathingSignal(g, Color.LightGreen); }   //【波形表示】デプス値　　　⇒『呼吸』

                            SolidBrush b = new SolidBrush(Color.White);
                            Font ft = new Font("Times New Roman", 30, FontStyle.Bold);
                            
                      //    g.DrawString("X=" + KS_ColorPoint[0].X.ToString() + ", Y=" + KS_ColorPoint[0].Y.ToString(), ft, b, new PointF(200, 400));
                      //    g.DrawString("Ir = " + Heart.ToString() + "", ft, b, new PointF(200, 360));
                            g.DrawString("心拍数 = " + HBR.ToString("F2") + "[拍/分]", ft, b, new PointF(150, 400));
                            g.DrawString("胸部距離 = " + BrValue.ToString("F3") + "[mm]", ft, b, new PointF(150, 440));
                            g.Dispose(); //【描画終了】
                            Validate();  //【更新を反映する】
                            break;
                        }
                    }
                }
            }
        }

        //***************************************************************
        //【GetIrAverage】心拍測定のために赤外線強度平均値を取得する
        //***************************************************************
        private float GetIrAverage(Rectangle r, int Vmax, int Vmin)
        {
            int X, Y, Z, k, L;
            long Sum = 0, Count = 0;
            for (Y = r.Top; Y < r.Top + r.Height; Y++)
            {
                L = KS_FrameWidth * Y;
                for (X = r.Left; X < r.Left + r.Width; X++)
                {
                    k = 2 * (X + L);
                    Z = KS_IrCamData[k + 1] * 256 + KS_IrCamData[k];
                    if ((KS_DepthData[X + L].PlayerIndex > 0) && (Z < Vmax) && (Z > Vmin))
                    {
                        Sum += Z;
                        Count++;
                    }
                }
            }
            return (float)Sum / (float)Count;
        }

        //***************************************************************
        //【GetAverage】呼吸測定のために距離平均値を取得する
        //***************************************************************
        private float GetAverage(Rectangle r, int Vmax, int Vmin)
        {
            int X, Y, Z, L;
            long Sum = 0, Count = 0;
            for (Y = r.Top; Y < r.Top + r.Height; Y++)
            {
                L = KS_FrameWidth * Y;
                for (X = r.Left; X < r.Left + r.Width; X++)
                {
                    Z = KS_DepthData[X + L].Depth;
                    if ((KS_DepthData[X + L].PlayerIndex > 0) && (Z < Vmax) && (Z > Vmin))
                    {
                        Sum += Z;
                        Count++;
                    }
                }
            }
            return (float)Sum / (float)Count;
        }

        //***************************************************************
        //【DisplayGrid】横２秒ピッチ、縦１ｍｍピッチグリッドを表示する
        //***************************************************************
        private void DisplayGrid(Graphics g, Color lc)
        {
            //【縦は10分割、デプスは10mmppに相当、赤外線脈波は100degit】
            int i;
            for (i = 0; i < 640; i += 60) //【時間軸は60で1秒】
            {
                g.DrawLine(new Pen(lc, 3.0f), i, 0, i, 480);       //【画面中央縦線】
            }
            g.DrawLine(new Pen(lc, 3.0f), 639, 0, 639, 480);       //【画面中央縦線】
            for (i = 0; i < 480; i += 48)
            {
                g.DrawLine(new Pen(lc, 3.0f), 0, i, 640, i);       //【画面中央横線】
            }
        }

        //***************************************************************
        //【DisplayBreathingSignal】FIFOクラスを使って波形を表示する
        //***************************************************************
        private void DisplayBreathingSignal(Graphics g, Color lc)
        {
            int i, j;
            double Dy;
            const int FrameHeight = 480;
            Dy = FrameHeight / 10.0; //【デプス呼吸は10mm_pp】
            for (i = 0, j = BrFIFO.Pointer; i < nFIFO; i++)
            {
                BrFIFO.Points[i].X = (int)i; //0～600
                BrFIFO.Points[i].Y = (int)((BrFIFO.dFIFO[j] - BrFIFO.OffsetLevel) * Dy) + 240; //【中央値をOffsetLevelとする】
                j++; if (j == nFIFO) { j = 0; } //【FIFO管理ポインタを更新する】
            }
            g.DrawLines(new Pen(lc, 5.0f), BrFIFO.Points);
        }

        //***************************************************************
        //【DisplayHeartRateSignal】FIFOクラスを使って波形を表示する
        //***************************************************************
        private void DisplayHeartRateSignal(Graphics g, Color lc)
        {
            int i, j;
            double Dy;
            const int FrameHeight = 480;
            Dy = FrameHeight / 100.0; //【赤外線脈波は100digits_pp】
            for (i = 0, j = IrFIFO.Pointer; i < nFIFO; i++)
            {
                IrFIFO.Points[i].X = (int)i; //0～600
                IrFIFO.Points[i].Y = (int)((IrFIFO.dFIFO[j] - IrFIFO.OffsetLevel) * Dy) + 240; //【中央値をOffsetLevelとする】
                j++; if (j == nFIFO) { j = 0; } //【FIFO管理ポインタを更新する】
            }
            g.DrawLines(new Pen(lc, 5.0f), IrFIFO.Points);
        }

    }
}
